import xml.etree.ElementTree as ElementTree
from netCDF4 import Dataset
from typing import Any
from uuid import uuid1
from xml.dom import minidom
from pathlib import Path
import re
import warnings


def add_nested_elements(top_element:str, element_type:str, add_attributes:"list[tuple[Any, dict]]") -> ElementTree.Element:
    _add_attributes = ElementTree.Element(top_element)
    
    for tup in add_attributes:
        _attr = ElementTree.SubElement(_add_attributes, element_type)
        _attr.text = str(tup[0])
        
        for key in tup[1].keys():
            _attr.set(str(key), str(tup[1][key]))
    
    return _add_attributes


class ErddapXmlGenerator:

    ERDDAP_BOOL = {True:'true', False:'false'}
    ERDDAP_DTYPE = {'float32': 'float',
                    'S1': 'String',
                    '|S1': 'String',
                    'str': "String",
                    str : "String",
                    "<class 'str'>":"String",
                    'uint64': 'int'}

    def __init__(self, datasetType="EDDTableFromNcCFFiles", datasetID=None, active=True):
        if not datasetID:
            datasetID = str(uuid1())
        
        self.xml = ElementTree.ElementTree(self.dataset(datasetType, datasetID, active))
        

    @classmethod
    def from_nc(cls,
                nc_file: str,
                datasetType="EDDTableFromNcCFFiles",
                datasetID=None, active=True,
                header_mods:"dict[str,str]"={},
                omit_variables:"list[str]"=[],
                copy_globals:"list[str]"=["cdm_altitude_proxy", "sourceUrl", "cdm_data_type", "featureType","subsetVariables","infoUrl"]):
        """_summary_

        Parameters
        ----------
        nc_file : str
            _description_
        datasetType : str, optional
            _description_, by default "EDDTableFromNcCFFiles"
        datasetID : _type_, optional
            _description_, by default None
        active : bool, optional
            _description_, by default True
        header_mods : dict[str,str], optional
            _description_, by default {}
        copy_globals : list[str], optional
            global attributes in the netcdf file to copy to the addAttributes, by default []

        Returns
        -------
        _type_
            _description_
        """
        
        with Dataset(nc_file) as rootgrp:
            has_id = hasattr(rootgrp, 'id')
            
            if datasetID:
                if has_id and id != datasetID:
                    warnings.warn(f"parameter datasetID ({datasetID}) differs from dataset id attribute ({id})")
            
            elif has_id:
                datasetID = rootgrp.getncattr('id')

            else:
                new_id = str(uuid1())
                warnings.warn(f"Missing id attribute. Auto-generating {str(uuid1())}")
                datasetID = new_id

        this = cls(datasetType, datasetID, active)
        
        this.writeHeaderDict(this.xml.getroot(), this.erddapHeaderDict(fileNameRegex=Path(nc_file).name))

        # modify header
        this.modify_xml(header_mods)

        # Write addAttributes for globals
        add_attrs = this.globals_from_nc(nc_file, copy_globals)
        this.xml.getroot().append(add_attrs)

        # Get data variable xml
        dv = this.dataset_variables_from_nc(nc_file)

        for e in dv:
            if e.find("sourceName").text in omit_variables:
                continue
            this.xml.getroot().append(e)

        return this

    @staticmethod
    def dataset_variables_from_nc(nc_file:str) -> "list[ElementTree.Element]":
        dv = list()
        with Dataset(nc_file, 'r') as rootgrp:
            for var in rootgrp.variables:
                dat_var = ErddapXmlGenerator.dataVariable(source_name=var,
                                                destination_name=var,
                                                data_type=ErddapXmlGenerator.get_dtype(str(rootgrp[var].dtype)))
                dv.append(dat_var)
        
        return dv

    def modify_xml(self, mods:"dict[str,str]"={}):
        for key, value in mods.items():
            
            if isinstance(value, bool):
                value = self.ERDDAP_BOOL[value]

            e = self.xml.getroot().find(key)
            
            if e.text:
                e.text = value  
    
    @staticmethod
    def globals_from_nc(nc_file:str, attrs:"list[str]") -> "ElementTree.Element":
        
        dv = list()
        with Dataset(nc_file, 'r') as rootgrp:
            for attr in rootgrp.ncattrs():
                if attr in attrs:
                    dv.append((attr, rootgrp.getncattr(attr)))
        
        return ErddapXmlGenerator.addAttributes(dv)


    @classmethod
    def from_csv(cls, csv: str):
        this = cls()
        return this
    
    def __add_essentials(self):
        root = self.xml.getroot()
        
        if not root.findall("att[@name='sourceUrl']"):
            if not root.find("addAttributes"):
                root.append(ElementTree.Element("addAttributes"))

            add_attributes = root.find("addAttributes")
            
            source_url = ElementTree.Element("att")
            source_url.set("name", "sourceUrl")
            source_url.text = "(local files)"
            add_attributes.append(source_url)

    def write_xml(self, file: str):
        self.__add_essentials()
        rough_string = ElementTree.tostring(self.xml.getroot()) 
        reparsed = minidom.parseString(rough_string)
        cleaned = reparsed.toprettyxml(indent="  ")
        no_header = re.sub(r"<\?xml .*\?>\n", "", cleaned)
        
        with open(file, 'wb') as f:
            f.write(no_header.encode('utf-8'))
                
    @staticmethod
    def get_dtype(dtype:str):
        return ErddapXmlGenerator.ERDDAP_DTYPE[dtype]

    @staticmethod
    def dataset(datasetType:str, datasetID:str, active:bool=True) -> ElementTree.Element:
        _dataset = ElementTree.Element("dataset")
        _dataset.set("type", datasetType)
        _dataset.set("datasetID", datasetID)
        _dataset.set("active", ErddapXmlGenerator.ERDDAP_BOOL[active])
        
        return _dataset

    @staticmethod
    def addAttributes(attrs:"list[tuple[str,str]]") -> ElementTree.Element:
        """Create XML addAttributes element with <att> children

        Parameters
        ----------
        attrs : list[tuple[str,str]]
            list of (key, value) e.g. [("sourceUrl", "www.myurl.com")]

        Returns
        -------
        ElementTree.Element
            _description_
        """
        a = [(text, {'name': attr}) for attr, text in attrs]
        e = add_nested_elements("addAttributes", "att", a)
        return e

    @staticmethod
    def dataVariable(source_name:str, destination_name:str, data_type:str, add_attributes:"list[tuple[Any, dict]]"=None) -> ElementTree.Element:
        e = ElementTree.Element("dataVariable")
        
        if source_name:
            _source_name = ElementTree.SubElement(e, "sourceName")
            _source_name.text = source_name
        
        if destination_name:
            _destination_name = ElementTree.SubElement(e, "destinationName")
            _destination_name.text = destination_name
        
        if data_type:
            _data_type = ElementTree.SubElement(e, "dataType")
            _data_type.text = data_type
        
        if add_attributes:
            e.append(ErddapXmlGenerator.addAttributes(add_attributes))
        
        return e
    
    @staticmethod
    def writeHeaderDict(element:ElementTree.Element, header_dict:dict) -> None:
        for key in header_dict:
            line = ElementTree.Element(key)
            value = header_dict[key] if header_dict[key] else ""
            line.text = str(value)
            
            element.append(line)
        
    @staticmethod
    def erddapHeaderDict(reloadEveryNMinutes=10080, updateEveryNMillis=10000,
                    fileDir="/", fileNameRegex=".*", recursive=False, pathRegex=".*", 
                    metadataFrom='last', standardizeWhat=0, sortFilesBySourceNames="",
                    fileTableInMemory=False, accessibleViaFiles=False) -> "dict[str,str]":
    
        _header = {"reloadEveryNMinutes": str(reloadEveryNMinutes), 
                  "updateEveryNMillis": str(updateEveryNMillis),
                  "fileDir": str(fileDir), 
                  "fileNameRegex": str(fileNameRegex), 
                  "recursive": ErddapXmlGenerator.ERDDAP_BOOL[recursive], 
                  "pathRegex": str(pathRegex), 
                  "metadataFrom": str(metadataFrom),
                  "standardizeWhat": str(standardizeWhat), 
                  "sortFilesBySourceNames": sortFilesBySourceNames,
                  "fileTableInMemory": ErddapXmlGenerator.ERDDAP_BOOL[fileTableInMemory],
                  "accessibleViaFiles": ErddapXmlGenerator.ERDDAP_BOOL[accessibleViaFiles]
                  }
        
        return _header

dataVariableDefaults = {
    ""
}


if __name__ == "__main__":
    E = ErddapXmlGenerator.from_nc(r"C:\Users\Nick\ESJ2.nc-pink_west_merged_clean1.nc", header_mods={"fileDir":"/cats/"})
    E.write_xml("C:/Users/Nick/xml2.xml")


import argparse
import netCDF4 as nc
import pandas as pd
import pathlib
import textwrap
from typing import Union

from pfit.CSVColMelter import CSVMelter
from pfit.pfnet_standard import make_multi_temperature_base, add_global_vars, calculate_extent_metadata, standard_atts

Filepath = Union[str, pathlib.Path]

class NtgsThermalDataset():
    """ Ground temperature data """
    def __init__(self, data_file, metadata, timezoneOffset = -7):
        self.data_file = data_file
        self.metadata = metadata
        self.timezoneOffset = timezoneOffset
        self.melted_data = self.read_NTGS(data_file)
    
    @classmethod
    def data_only(cls, data_file, timezoneOffset=-7):
        this = cls(data_file, timezoneOffset)
        return this

    @staticmethod
    def dirExists(path):
        if pathlib.Path(path).is_dir():
            return path
        else:
            raise argparse.ArgumentTypeError(f"The directory specified does not exist: {path}")

    def read_NTGS(self, ntgs_file):
        melter = CSVMelter(self.timezoneOffset)
        return melter.getMeltedDataframe(ntgs_file)

    def to_netcdf(self, nc_file:"Filepath"=None) -> str:
        """ Write dataset to a netcdf file """
        melted_df = self.melted_data 
        
        info = self.extract_info(melted_df)
        if nc_file is None:
            nc_file = pathlib.Path(self.data_file).with_suffix(".nc")
        else:
            nc_file = pathlib.Path(nc_file).with_suffix(".nc")

        rootgroup = self.make_ntgs_nc(nc_file,
                                     depths=info['depths'],
                                     times=info['times'],
                                     borehole=info['site_name'],
                                     data=info['data'],
                                     latitude=info['latitude'],
                                     longitude=info['longitude'])
        
        return str(nc_file)

    @staticmethod 
    def make_ntgs_nc(nc_target, depths, times, borehole, data, latitude: float, longitude: float) -> None:
        make_multi_temperature_base(nc_target, len(depths))
        
        with nc.Dataset(nc_target, 'a') as rootgrp:
        
            add_global_vars(rootgrp, standard_atts)
            rootgrp['latitude'][:] = latitude
            rootgrp['longitude'][:] = longitude
            rootgrp['depth'][:] = depths
            rootgrp['site_name'][:] = nc.stringtoarr(borehole, rootgrp['site_name'].shape[1])
            rootgrp['time'][:] = nc.date2num(times, rootgrp['time'].units)
            rootgrp['ground_temperature'][0, :, :] = data
            
            calculate_extent_metadata(rootgrp)
        
        return nc_target

    @staticmethod
    def extract_info(melted_df: pd.DataFrame) -> dict:
        """ Get data out of melted dataframe """
        info = dict()
        
        data = melted_df.pivot(index='time', columns='depth', values='temperature')
        data.replace(r"^\s*$", "nan", regex=True, inplace=True)  # TODO: make sure this is cleaned earlier

        info['data'] = data
        info['site_name'] = melted_df['site_id'][0]
        info['latitude'] = melted_df['latitude'][0]
        info['longitude'] = melted_df['longitude'][0]
        info['depths'] = data.columns.to_numpy()
        info['times'] = pd.to_datetime(data.index).to_pydatetime()
        info['n_depths'] = len(melted_df['depth'].unique())

        return info


# def nc_from_melted(nc_target, melted_file, metadata, metadata_id=None):
    
#     melted = pd.read_csv(melted_file)
#     data = melted.pivot(index='time', columns='depth', values='temperature')
    
#     if metadata_id is None:
#         metadata_id = melted["site_id"]

#     depths = data.columns.to_numpy()
#     times = pd.to_datetime(data.index)

#     borehole = melted['site_id'][0]
#     latitude = melted['latitude'][0]
#     longitude = melted['longitude'][0]
#     make_ntgs_nc(nc_target, depths, times, borehole, data,  latitude, longitude)



if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''
        Converts CSV, XLS, or XLSX files in the NTGS-style "wide" format to NetCDF files.
        The file being transformed must be free of errors prior to running this.
        '''))
    parser.add_argument('location', metavar='path', type=CSVMelter.pathExists, help='Path to input dataset file')
    parser.add_argument('output-path', metavar='output-path', type=NtgsThermalDataset.dirExists, help='Directory where the NetCDF file will be written to')
    parser.add_argument('--timezone-offset', metavar='timezone-offset', type=CSVMelter.timezone_check, help='Optional value indicating timezone offset from UTC-0 (default: UTC-7). To represent half hour, use "-3.5" for Newfoundland time as an example.')
    arguments = vars(parser.parse_args())
    dataset = None
    if arguments['timezone_offset'] is not None:
        dataset = NtgsThermalDataset(arguments['location'], None, arguments['timezone_offset'])
    else:
        dataset = NtgsThermalDataset(arguments['location'], None)
    dataset.to_netcdf(arguments['output-path'])



from pfit.MetadataHandler import MetadataHandler
from netCDF4 import Dataset, chartostring
import pandas as pd
from numpy import atleast_1d
from typing import Type, Union
import re

class TabularMetadataHandler(MetadataHandler):
    """ 
    index_column : str or int
        index of the index column (starting from 1) or name of column with unique records
        corresponding to which file or site to write to
        
    column names mini-language: Use these patterns to write your metadata attributes
    "att_name" create global attribute with name "att_name"
    "!var:dest" create attribute "dest" for variable "var"
    "![@att='val']:dest" create attribute "dest" for all variables where attribute "att" has value "val"
    "!$var" write value to a single-value scalar variable "var"
    "!$[@att='val'] write value to single value scalar variable which has attribute "att" with value "val"
    """
    R_SIMPLE_VARIABLE = re.compile(r"^!(\w*):(\w*)$")
    R_SCALAR_VALUE = re.compile(r"^!\$(\w*)$")

    def __init__(self, table: pd.DataFrame, index_column: Union[int, str]):
        self.table=table
        self.index = index_column
        self.table = self.table.set_index(self.index)
        self.tabledict = self.table.to_dict("index")

    @classmethod
    def from_csv(cls):
        pass

    @classmethod
    def from_xlsx(cls, xl_file, sheet_index: int, index_column: Union[int, str], **kwargs):
        xl_data = pd.read_excel(xl_file, sheet_name=sheet_index, skiprows=[1], engine='openpyxl', **kwargs)
        this = cls(xl_data, index_column)
        return this

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value):
        if isinstance(value, int):
            value = self.table.columns[1 + value]
        
        self._index= value

    def write_variables(self, nc_file: Dataset, loc:str) -> None:
        pass

    def _lookup_loc(self, ncDataset: Dataset, lookup) -> str:
        locations = atleast_1d(ncDataset[lookup][:])
        
        if locations.dtype == "S1": 
            sites = chartostring(locations)

        elif hasattr(ncDataset[lookup], "_Encoding"):
            sites = locations

        else:
            raise TypeError(f"Variable {lookup} is not a string or cannot be read")

        if len(sites) != 1:
            raise ValueError("too many sites returned. Only works for a dataset with one site")

        return sites[0]

    def write_attributes(self, nc_file: str, loc:str=None, loc_lookup:str=None) -> None:
        """ 
        loc_name : str
            Identifier to match against index_column. ignored if loc_lookup is provided
        loc_lookup : str
            the name of a variable in nc_file that contains strings matching strings in 
            the defined index_column """
        if (not loc and not loc_lookup):
            raise ValueError("one of 'loc' or 'loc_lookup' must be provided")
        
        with Dataset(nc_file, 'a') as rootgrp:
            if loc_lookup:
                loc = self._lookup_loc(rootgrp, loc_lookup)

            for key in self.get_attributes(loc):
                value = self.get_attribute(loc, key) 
                
                if not self._is_special_attribute(key):
                    rootgrp.setncattr(key, value)
                
                elif self.R_SIMPLE_VARIABLE.match(key):
                    var, att = self._get_simple_variable_assignment(key)
                    rootgrp[var].setncattr(att, value)
                
                elif self.R_SCALAR_VALUE.match(key):
                    scalar_variable = self._get_scalar_value_assignment(key)
                    rootgrp[scalar_variable][:] = value

                else:
                    pass
    
    def _get_scalar_value_assignment(self, string:str) -> str:
        match = self.R_SCALAR_VALUE.match(string)

        if match:
            return match.group(1)
        else:
            return ""

    def _get_simple_variable_assignment(self, string: str) -> tuple:
        match = self.R_SIMPLE_VARIABLE.match(string)
        
        if match:
            assignment = (match.group(1), match.group(2))
            return assignment
        else:
            return (None, None)

    def _is_special_attribute(self, name: str) -> bool:
        special_chars = ["!", "$", "@"]
        is_special = any([x in name for x in special_chars])

        return is_special

    def get_attributes(self, loc: str) -> dict:
        return self.tabledict[loc]

    def get_attribute(self, loc: str, attribute: str) -> str:
        return self.tabledict[loc][attribute]

if __name__ == "__main__":
    t = TabularMetadataHandler.from_xlsx(r"C:/Users/Nick/Downloads/ESJ_Metadata.xlsx", 0, "platform_name")


import netCDF4 as nc
import pandas as pd
from pfit.ntgs_metdata import ACDD_globals, ACDD_variable_attributes, ntgs_global_default
from pfit.variables_templates import add_variable
from numpy import diff
from scipy.stats import mode

#    GCW / WMO recommends including only one borehole per dataset (erddap/opendap)
#    but this means that if you have a larger dataset that comprises many boreholes, 
#    each will be required to have its own title & ID, whereas you might want the files to be 
#    grouped into their own title, and especially if "ID" is used to be the DOI, you want one for
#    the larger datset.
#
#    Ground temperature stations are much less costly and easier to set up than ocean mooring cables,
#    so it is more likely that a dataset submission will have many more of them.
#     
#    Including multiple stations in a file would require having indices for depth and time, because
#    not all stations will have the same depths (or number of sensors), and they may not
#    record at the same time (and may not have the same number of recordings).
#
#    One solution is to not expect that individual sites have *all* of the ACDD information since
#    we are not using them to represent 'full' datasets. This keeps them useful as a way to store and
#    transmit data. If we want to combine them into 'full' datasets with IDs then that is possible with
#    some further work.


def make_temperature_base(file, ndepth):

    with nc.Dataset(file, 'w', format='NETCDF4') as rootgrp: 

        # Create dimensions
        _  = rootgrp.createDimension('depth_below_ground_surface', ndepth)
        _  = rootgrp.createDimension('time', None)
        _  = rootgrp.createDimension('nchar32', 32)

        # Create coordinate variables
        Lat = rootgrp.createVariable('latitude', 'f4', )
        Lat.standard_name = "latitude"
        Lat.long_name = "latitude"
        Lat.units = "degrees_north"
        Lat.axis  = "Y"
        Lat.ioos_category = "Location"

        Lon = rootgrp.createVariable('longitude', 'f4', )
        Lon.standard_name = "longitude"
        Lon.long_name = "longitude"
        Lon.units = "degrees_east"
        Lon.axis  = "X"
        Lon.ioos_category = "Location"

        Elev = rootgrp.createVariable('surface_elevation', 'f4')
        # Elev.standard_name = "height_above_reference_ellipsoid"
        Elev.long_name = "Ground surface elevation"
        Elev.units = "m"
        Elev.ioos_category = "Location"

        Z = rootgrp.createVariable('depth_below_ground_surface', 'f4', ('depth_below_ground_surface',))
        Z.standard_name = "depth"
        Z.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Z.long_name = "Depth to observation"
        Z.units = "m"
        Z.axis  = "Z"
        Z.positive = "down"
        Z.ioos_category = "Location"

        Time = rootgrp.createVariable('time', 'u8', ('time',))
        Time.standard_name = "time"
        Time.cf_role = 'profile_id'
        Time.units = "seconds since 1900-01-01"
        Time.calendar = "gregorian"
        Time.axis  = "T"
        Time.ioos_category = "Time"

        Profile = rootgrp.createVariable('site_name', 'S1', ('nchar32'))
        Profile.standard_name = "platform_name"
        Profile._Encoding = "UTF-8"
        Profile.long_name = "Borehole or site name"
        Profile.cf_role  = "timeseries_id"
        Profile.ioos_category = "Identifier"    

        T = rootgrp.createVariable('ground_temperature', 'f4', ('time', 'depth_below_ground_surface'))
        T.standard_name = "temperature_in_ground"
        T.standard_name_url = "https://mmisw.org/ont/cf/parameter/temperature_in_ground"
        T.units = "degree_C"
        T.coordinates = "site_name depth_below_ground_surface latitude longitude"
        T.long_name = "Ground temperature"
        T.platform = "site_name"
        T.ioos_category = "Temperature"

        rootgrp.cdm_altitude_proxy = "depth_below_ground_surface"
        
        for key, value in standard_atts.items():
            rootgrp.setncattr(key, value)

    return file

standard_atts = {
    'featureType': 'timeSeriesProfile',
    'Conventions': "COARDS, CF-1.8, ACDD-1.3",
    'cdm_profile_variables': "time",
    "standard_name_vocabulary": "CF Standard Name Table v78"
}


def make_multi_temperature_base(file, ndepth):

    with nc.Dataset(file, 'w', format='NETCDF4') as rootgrp: 

        # Create dimensions
        _  = rootgrp.createDimension('station', 1)
        _  = rootgrp.createDimension('z', ndepth)
        _  = rootgrp.createDimension('time', None)
        _  = rootgrp.createDimension('nchar32', 32)

        # Create coordinate variables
        Lat = rootgrp.createVariable('latitude', 'f4', ('station',))
        Lat.standard_name = "latitude"
        Lat.long_name = "latitude"
        Lat.units = "degrees_north"
        Lat.axis  = "Y"
        Lat.ioos_category = "Location"

        Lon = rootgrp.createVariable('longitude', 'f4', ('station',))
        Lon.standard_name = "longitude"
        Lon.long_name = "longitude"
        Lon.units = "degrees_east"
        Lon.axis  = "X"
        Lon.ioos_category = "Location"

        Elev = rootgrp.createVariable('elevation', 'f4', ('station',))
        Elev.standard_name = "height_above_reference_ellipsoid"
        Elev.units = "m"
        Elev.ioos_category = "Location"

        Time = rootgrp.createVariable('time', 'u8', ('time',))
        Time.standard_name = "time"
        Time.cf_role = 'profile_id'
        Time.units = "seconds since 1900-01-01"
        Time.calendar = "gregorian"
        Time.axis  = "T"
        Time.ioos_category = "Time"

        Profile = rootgrp.createVariable('site_name', 'S1', ('station', 'nchar32'))
        Profile.standard_name = "platform_name"
        Profile._Encoding = "UTF-8"
        Profile.long_name = "Borehole or site name"
        Profile.cf_role  = "timeseries_id"
        Profile.ioos_category = "Identifier"    

        T = rootgrp.createVariable('ground_temperature', 'f4', ('station', 'time', 'z'))
        T.standard_name = "temperature_in_ground"
        T.standard_name_url = "https://mmisw.org/ont/cf/parameter/temperature_in_ground"
        T.units = "degree_C"
        T.coordinates = "site_name depth latitude longitude"
        T.long_name = "Ground temperature"
        T.platform = "site_name"
        T.ioos_category = "Temperature"

        rootgrp.featureType = 'timeSeriesProfile' 
        rootgrp.cdm_altitude_proxy = "depth_below_ground_surface"
        
    return file


def add_global_vars(rootgrp, att_dict):
    for key in att_dict:
        rootgrp.setncattr(key, att_dict[key])

def add_variables(rootgrp, var_dict):
    for key in var_dict:
        var = var_dict[key]
        new_variable = rootgrp.createVariable(key, var['type'], var['dimensions'])
        for attr, value in var['attributes'].items():
            new_variable.setncattr(attr, value)

def write_variables(rootgrp, dataframe, var_dict):
    for key in var_dict:
        
        if rootgrp[key]._isvlen:
            for i in range(dataframe[key]):
                rootgrp[key][i] = dataframe[key][i]
        
        else:
            rootgrp[key][:] = dataframe[key]

def calculate_extent_metadata(rootgrp):
    """ calculate extents from data. Data variables must be populated"""
    t_units = rootgrp['time'].units
    min_date = nc.num2date(min(rootgrp['time'][:]), t_units)
    max_date = nc.num2date(max(rootgrp['time'][:]), t_units)

    rootgrp.time_coverage_start = min_date.isoformat()
    rootgrp.time_coverage_end = max_date.isoformat()
    rootgrp.time_coverage_duration = pd.Timedelta(max_date - min_date).isoformat()
    rootgrp.time_coverage_resolution = pd.Timedelta(mode(diff(nc.num2date(rootgrp['time'][:], t_units)), nan_policy='omit').mode[0]).isoformat()
    rootgrp.geospatial_bounds = f"POINT({rootgrp['longitude'][:][0]} {rootgrp['latitude'][:][0]})"

    rootgrp.geospatial_lat_min = min(rootgrp['latitude'][:])
    rootgrp.geospatial_lat_max = max(rootgrp['latitude'][:])
    rootgrp.geospatial_lon_min = min(rootgrp['longitude'][:])
    rootgrp.geospatial_lon_max = max(rootgrp['longitude'][:])

    rootgrp.observation_depth_min = min(abs(rootgrp['depth'][:]))
    rootgrp.observation_depth_max = max(abs(rootgrp['depth'][:]))

    rootgrp.geospatial_vertical_max = min(rootgrp['elevation'][:])
    rootgrp.geospatial_vertical_max = max(rootgrp['elevation'][:])



def make_profile_base(file):

    with nc.Dataset(file, 'w', format='NETCDF4') as rootgrp: 

        # Create dimensions
        _ = rootgrp.createDimension('depth_below_ground_surface', None)
        _ = rootgrp.createDimension('nchar64', 64)
        _ = rootgrp.createDimension('nbnd', 2)

        # Create coordinate variables
        Lat = rootgrp.createVariable('latitude', 'f4', ())
        Lat.standard_name = "latitude"
        Lat.standard_name_url = "https://mmisw.org/ont/cf/parameter/latitude"
        Lat.long_name = "latitude"
        Lat.units = "degrees_north"
        Lat.axis  = "Y"
        Lat.ioos_category = "Location"

        Lon = rootgrp.createVariable('longitude', 'f4', ())
        Lon.standard_name = "longitude"
        Lon.standard_name_url = "https://mmisw.org/ont/cf/parameter/longitude"
        Lon.long_name = "longitude"
        Lon.units = "degrees_east"
        Lon.axis  = "X"
        Lon.ioos_category = "Location"
        
        Profile = rootgrp.createVariable('profile', 'S1', ('nchar64'))
        Profile.cf_role = 'profile_id'
        Profile._Encoding = "UTF-8"
        Profile.long_name = "Unique profile identifier"
        Profile.ioos_category = "Identifier"

        Elev = rootgrp.createVariable('surface_elevation', 'f4', ())
        Elev.long_name = "Ground surface elevation"
        Elev.standard_name = "height_above_reference_ellipsoid"
        Elev.units = "m"
        Elev.ioos_category = "Location"

        Time = rootgrp.createVariable('time', 'u8', ())
        Time.long_name = "Date of measurement"
        Time.standard_name_url = "https://mmisw.org/ont/cf/parameter/time"
        Time.standard_name = "time"
        Time.units = "seconds since 1900-01-01"
        Time.calendar = "standard"
        Time.axis  = "T"
        Time.ioos_category = "Time"

        Name = rootgrp.createVariable('platform_name', 'S1', ('nchar64'))
        Name.standard_name = "platform_name"
        Name.long_name = "Borehole or site name"
        Name.ioos_category = "Identifier"
        Name._Encoding = "UTF-8"

        Z = rootgrp.createVariable('depth_below_ground_surface', 'f4', ('depth_below_ground_surface',))
        Z.standard_name = "depth"
        Z.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Z.long_name = "Depth to midpoint of interval"
        Z.units = "m"
        Z.axis  = "Z"
        Z.positive = "down"
        Z.ioos_category = "Location"
        Z.bounds = "depth_bounds"

        Zbnd = rootgrp.createVariable('depth_bounds', 'f4', ('depth_below_ground_surface','nbnd'))
        Zbnd.long_name = "Depth below ground surface"
        Zbnd.ioos_category = "Location"

        Ztop = rootgrp.createVariable('top_of_interval', 'f4', ('depth_below_ground_surface',))
        Ztop.standard_name = "depth"
        Ztop.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Ztop.long_name = "Depth to top of interval"
        Ztop.units = "m"
        Ztop.positive = "down"
        Ztop.ioos_category = "Location"

        Zbot = rootgrp.createVariable('bottom_of_interval', 'f4', ('depth_below_ground_surface',))
        Zbot.standard_name = "depth"
        Zbot.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Zbot.long_name = "Depth to bottom of interval"
        Zbot.units = "m"
        Zbot.positive = "down"
        Zbot.ioos_category = "Location"

        rootgrp.featureType = 'profile'
        rootgrp.cdm_data_type = "profile"
        rootgrp.cdm_altitude_proxy = "depth_below_ground_surface"
        rootgrp.Conventions = "CF-1.6, ACDD-1.3"
        rootgrp.cdm_profile_variables = "platform_name,time,latitude,longitude,surface_elevation"
        rootgrp.standard_name_vocabulary = "CF Standard Name Table v78"

        
    return file


def make_multi_profile_base(file):

    with nc.Dataset(file, 'w', format='NETCDF4') as rootgrp: 

        # Create dimensions
        _ = rootgrp.createDimension('z', None)
        _ = rootgrp.createDimension('station', None)
        _ = rootgrp.createDimension('nchar64', 64)
        _ = rootgrp.createDimension('nbnd', 2)

        # Create coordinate variables
        Lat = rootgrp.createVariable('latitude', 'f4', ('station',))
        Lat.standard_name = "latitude"
        Lat.standard_name_url = "https://mmisw.org/ont/cf/parameter/latitude"
        Lat.long_name = "latitude"
        Lat.units = "degrees_north"
        Lat.axis  = "Y"
        Lat.ioos_category = "Location"

        Lon = rootgrp.createVariable('longitude', 'f4', ('station',))
        Lon.standard_name = "longitude"
        Lon.standard_name_url = "https://mmisw.org/ont/cf/parameter/longitude"
        Lon.long_name = "longitude"
        Lon.units = "degrees_east"
        Lon.axis  = "X"
        Lon.ioos_category = "Location"
        
        Profile = rootgrp.createVariable('profile', 'S1', ('station','nchar64'))
        Profile.cf_role = 'profile_id'
        Profile._Encoding = "UTF-8"
        Profile.long_name = "Unique profile identifier"
        Profile.ioos_category = "Identifier"

        Elev = rootgrp.createVariable('surface_elevation', 'f4', ('station',))
        Elev.long_name = "Ground surface elevation"
        Elev.standard_name = "height_above_reference_ellipsoid"
        Elev.units = "m"
        Elev.ioos_category = "Location"

        Time = rootgrp.createVariable('time', 'u8', ('station',))
        Time.long_name = "Date of measurement"
        Time.standard_name_url = "https://mmisw.org/ont/cf/parameter/time"
        Time.standard_name = "time"
        Time.units = "seconds since 1900-01-01"
        Time.calendar = "standard"
        Time.axis  = "T"
        Time.ioos_category = "Time"

        Name = rootgrp.createVariable('platform_name', 'S1', ('station','nchar64'))
        Name.standard_name = "platform_name"
        Name.long_name = "Borehole or site name"
        Name.ioos_category = "Identifier"
        Name._Encoding = "UTF-8"

        Z = rootgrp.createVariable('depth_below_ground_surface', 'f4', ('station','z',))
        Z.standard_name = "depth"
        Z.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Z.long_name = "Depth to midpoint of interval"
        Z.units = "m"
        Z.axis  = "Z"
        Z.positive = "down"
        Z.ioos_category = "Location"
        Z.bounds = "depth_bounds"

        Zbnd = rootgrp.createVariable('depth_bounds', 'f4', ('station','z','nbnd'))
        Zbnd.long_name = "Depth below ground surface"
        Zbnd.ioos_category = "Location"

        Ztop = rootgrp.createVariable('top_of_interval', 'f4', ('station','z',))
        Ztop.standard_name = "depth"
        Ztop.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Ztop.long_name = "Depth to top of interval"
        Ztop.units = "m"
        Ztop.positive = "down"
        Ztop.ioos_category = "Location"

        Zbot = rootgrp.createVariable('bottom_of_interval', 'f4', ('station','z',))
        Zbot.standard_name = "depth"
        Zbot.standard_name_url = "https://mmisw.org/ont/cf/parameter/depth"
        Zbot.long_name = "Depth to bottom of interval"
        Zbot.units = "m"
        Zbot.positive = "down"
        Zbot.ioos_category = "Location"

        rootgrp.featureType = 'profile'
        rootgrp.cdm_data_type = "profile"
        rootgrp.cdm_altitude_proxy = "depth_below_ground_surface"
        rootgrp.Conventions = "CF-1.6, ACDD-1.3"
        rootgrp.cdm_profile_variables = "platform_name,time,latitude,longitude,surface_elevation"
        rootgrp.standard_name_vocabulary = "CF Standard Name Table v78"

        
    return file

if __name__ == "__main__":
    x  =make_profile_base(r"C:/tmp/ncprofile2.nc")
    


import requests
import xml.etree.ElementTree as ET
from typing import Union, Optional

class CFNames:
    """ An object for accessing CF Standard Name information
    version : int
        version number of the standard names table to access
    """
    XML_URL = "https://cfconventions.org/Data/cf-standard-names/{version}/src/cf-standard-name-table.xml"

    def __init__(self, version: int):
        self.__cf_version = version
        self.names = self._retrieve_data(version)

    def _retrieve_data(self, version: int) -> ET.Element:
        if not isinstance(version, int):
            raise TypeError("version must be an integer")

        request_url = self._get_xml_url_for_version(version)
        response = requests.get(request_url)

        if response.status_code == 404:
            raise KeyError(f"Version {version} not found")

        cf_xml = ET.fromstring(response.content)
        
        return cf_xml

    def _get_xml_url_for_version(self, version: int) -> str:
        return self.XML_URL.format(version=version)

    def is_standard_name(self, name: str) -> bool:
        try:
            name_xml = self._get_standard_name(name)
            return True
        
        except KeyError:
            return False

    def get_description(self, name: str) -> Optional[str]:
        try:
            name_xml = self._get_standard_name(name)
        except KeyError:
            raise KeyError(f"Could not retrieve standard_name: {name}")
        
        description = name_xml.find("./description").text
        
        return description
    
    def get_units(self, name: str) -> str:
        try:
            name_xml = self._get_standard_name(name)
        except KeyError:
            raise KeyError(f"Could not retrieve standard_name: {name}")

        units = name.find("./canonical_units").text
        
        return units

    def get_version(self) -> int:
        return self.__cf_version

    def _get_standard_name(self, name: str, allow_alias: bool=True) -> ET.Element:
        standard_name = None
        try:
            standard_name = self.__get_standard_name(name)
        
        except KeyError:
            if allow_alias:
                try:
                    standard_name = self.__get_standard_name_by_alias(name)
                except:
                    pass
        
        if standard_name is None:
            raise KeyError("No entry by that name")

        return standard_name
            
    def __get_standard_name(self, name: str) -> Optional[ET.Element]:
        element = self.names.find(f"entry[@id='{name.lower()}']")
        
        if element == -1:
            raise KeyError("No entry by that name")
        
        return element

    def __get_standard_name_by_alias(self, name: str) -> Optional[ET.Element]:
        alias = self.names.find(f"alias[@id='{name.lower()}']")
        
        if alias == -1:
            raise KeyError("No alias by that name")
        
        standard_name = alias.find("entry_id").text
        element = self._get_standard_name(standard_name)
        
        return element



import netCDF4 as nc
import functools
import re
from pprint import pprint

REQUIRED = ['summary', 'title', 'infoUrl', 'institution']

CONDITIONAL = {  # If A=B then C required
        ("cdm_data_type", "TimeSeriesProfile", "cdm_profile_variables"),
        ("cdm_data_type", "profile", "cdm_profile_variables")
}

BAD_MATCHES = [
    ("id", r"\s+", "Spaces in id field")
]

CODES = [
    (200, "Could not test"),
    (100, "Missing required attribute"),
    (101, "Problem with attribute value"),
    (102, "Required attribute is empty"),
    (0, "Passed"),
    (-1, "Not applicable"),
]


def is_error(check: "tuple[int, str]") -> bool:
    return 100 <= check[0] < 200


def is_warning(check: "tuple[int, str]") -> bool:
    return 200 <= check[0] < 300


def is_ok(check: "tuple[int, str]") -> bool:
    return 0 <= check[0] <100

def is_n_a(check: "tuple[int, str]") -> bool:
    return check[0] == -1


class Result:

    def __init__(self, checks: "list[tuple]"):
        self.checks = checks

    @property
    def errors(self) -> "list[str]":
        return [check[1] for check in self.checks if is_error(check)]

    @property
    def warnings(self) -> "list[str]":
        return [check[1] for check in self.checks if is_warning(check)]

    @property
    def passes(self) -> "list[str]":
        return [check[1] for check in self.checks if is_ok(check)]

    @property
    def skipped(self) -> "list[str]":
        return [check[1] for check in self.checks if is_n_a(check)]

    def summary(self):
        return((f"{len(self.checks)} checks performed:\n{len(self.errors)} Errors\n{len(self.warnings)} Warnings\n"
                f"{len(self.passes)} Passed checks\n{len(self.skipped)} Skipped checks"))

    def print_errors(self):
        pprint(self.errors)
    
    def print_warnings(self):
        pprint(self.warnings)


@functools.singledispatch
def check_erd_compatibility(ncdf: "nc.Dataset", silent=False) -> Result:
    """Check file compatibilty for ERDDAP

    Parameters
    ----------
    ncdf : nc.Dataset
        netCDF file to check

    Returns
    -------
    Result
        Result object
    """
    checks = []

    for attr in REQUIRED:
        checks.append(check_attribute(ncdf, attr))
    
    # has conditional metadata
    for (conditional, value, required) in CONDITIONAL:
        check = check_conditional_attribute(ncdf, conditional, value, required)
        checks.append(check) 
    
    # Problems with fields
    for (attr, expr, message) in BAD_MATCHES:
        checks.append(check_field_values(ncdf, attr, re.compile(expr), message))

    result = Result(checks)
    
    if not silent:
        print(result.summary())
    
    return result
    

@check_erd_compatibility.register
def _(ncdf: str, *args, **kwargs) -> Result:
    with nc.Dataset(ncdf) as filepath:
        return check_erd_compatibility(filepath, *args, **kwargs)


def check_attribute(ncdf: "nc.Dataset", attr_name: "str") -> "tuple[int, str]":
    if not hasattr(ncdf, attr_name):
        return (100, f"Missing attribute: {attr_name}")
    
    if getattr(ncdf, attr_name) in [None, ""]:
         return (102, f"Attribute {attr_name} must not be blank")
         
    else:
        return (0, "")


def check_field_values(ncdf, attr, pattern: "re.Pattern", message) -> "tuple[int, str]":
    if not hasattr(ncdf, attr):
        return (200, f"Could not test. Missing attribute {attr}")

    if pattern.search(getattr(ncdf, attr)):
        return (101, message)
    
    else:
        return (0, "")

def check_conditional_attribute(ncdf: "nc.Dataset", conditional_attr: "str", conditional_value: "str", required_attr: "str") -> "tuple[int, str]":
    
    if not hasattr(ncdf, conditional_attr):
        return (200, f"Missing attribute {conditional_attr}. Could not test {required_attr}.")

    check_value = getattr(ncdf, conditional_attr)

    if check_value != conditional_value:
        return (-1, "")

    if hasattr(ncdf, required_attr):
        return (0, "")
    
    else:
        return (100, f"Missing attribute {required_attr}. (Because {conditional_attr} == {conditional_value}")


def main():
    result = check_erd_compatibility(r"C:\Users\Nick\Downloads\noodle.nc")
    result.print_errors()


if __name__ == "__main__":
    main()







# Check for NTGS standard of files.
#
# *Only checks the first sheet of a file

import argparse
import datetime
import textwrap
import logging
import numpy as np
import pandas as pd
import pathlib
import re
import sys
import time
import zipfile as zip

class FileDataChecker():
    # Index offset: CSV, Excel files contain column headers in first row, plus zero-indexed Python
    I_OFFSET = 2

    # Time regex for HH:MM:SS but with proper values (no 99 seconds, 25 hours, etc.)
    TIME_RGX = re.compile('^(?:2[0-3]|[01]?[0-9]):[0-5][0-9]:[0-5][0-9]')

    # Metre match - should expect all other columns to be #.#_m
    METRE_RGX = re.compile('^(?:(?!.*)|[+-]?((\d+\.?\d*)|(\.\d+)))(_m)$')

    # Error Dictionary
    ERRORS = {
        'time': 'Time value missing or in an invalid format: ',
        'date': 'Date value missing or in an invalid format: ',
        'unexpectedColumn': 'An expected column name was found in an incorrect position: ',
        'extraColumn': 'An extra column not part of the NTGS standard was found and ignored (this may cause unexpected behaviour where a NTGS-formatted file is expected): ',
        'unsupportedExt': 'The following file type is not supported: ',
        'coordinate': 'Latitude or longitude value could not be read or is formatted incorrectly: ',
        'latitude': 'Latitude value is not valid: ',
        'longitude': 'Longitude value is not valid: ',
        'temperature': 'Temperature value could not be read or is formatted incorrectly: ',
        'noMeasures': 'No columns containing temperature data were found.',
        'missingColumn': 'One or more of the expected columns was not found in the table.  Make sure all columns from the template are included.',
        'nonUniqueTime': 'Duplicate timestamp found: '
    }

    FILETYPES = ["text", "excel"]

    # Expected column names in the exact order
    EXPECTED_COLS = [
        'project_name',
        'site_id',
        'latitude',
        'longitude',
        'date_YYYY-MM-DD',
        'time_HH:MM:SS'
    ]
    
    def __init__(self):
        self._filetype=None
        self.log = logging.getLogger('main')

    def collectErrorIndex(self, errString, errValue, index):
        self.log.info(errString + str(errValue) + ' @ ROW ' + str((index + self.I_OFFSET)))

    def collectError(self, errString, errValue):
        self.log.info(errString + str(errValue))

    def checkTimeFmt(self, time, index):
        if self.TIME_RGX.match(str(time)) is None:
            self.collectErrorIndex(self.ERRORS['time'], time, index)

    def checkDateFmt(self, date, index):       
        try:
            datetime.datetime.strptime(str(date), "%Y-%m-%d")
        except ValueError:
            self.collectErrorIndex(self.ERRORS['date'], date, index)

    def checkExtraColNames(self, columns):
        extra = []
        for columnName in columns:
            if columnName not in self.EXPECTED_COLS:
                if self.METRE_RGX.match(columnName) is None:
                    extra.append(columnName)
                    self.collectError(self.ERRORS['extraColumn'], columnName)
        return extra

    def checkColNames(self, cols):
        for i, name in enumerate(cols):
            if (i < len(self.EXPECTED_COLS)):
                if (name != self.EXPECTED_COLS[i]):
                    # If the current column name is not in the exact position with the exact expected name
                    self.collectError(self.ERRORS['unexpectedColumn'], name)

    def checkTemperatures(self, temp, index):
        try:
            if temp == "":
                # Blank temperature values are OK
                return
            else:
                val = float(temp)
                if np.isnan(val):
                    # If it was a NaN (which is a valid float after conversion) - it's not good
                    self.collectErrorIndex(self.ERRORS['temperature'], temp, index)
        except ValueError:
            # Failed to convert to float
            self.collectErrorIndex(self.ERRORS['temperature'], temp, index)
      
    def checkLatitude(self, lat, index):
        # Only checks if the values here are valid latitude values, not if they have drifted off or are inconsistent.
        # They are expected to be the same throughout (per station/site).
        conversionFailure = False
        
        try:
            lat = float(lat)
        except ValueError:
            conversionFailure = True
            self.collectErrorIndex(self.ERRORS['coordinate'], lat, index)
        
        if conversionFailure: return
            
        if lat < -90 or lat > 90:
            self.collectErrorIndex(self.ERRORS['latitude'], lat, index)
            
    def checkLongitude(self, longitude, index):
        # Only checks if the values here are valid longitude values, not if they have drifted off or are inconsistent.
        # They are expected to be the same throughout (per station/site).
        conversionFailure = False
        
        try:
            longitude = float(longitude)
        except ValueError:
            conversionFailure = True
            self.collectErrorIndex(self.ERRORS['coordinate'], longitude, index)
        
        if conversionFailure: return
            
        if longitude < -180 or longitude > 180:
            self.collectErrorIndex(self.ERRORS['longitude'], longitude, index)
    
    def checkUniqueTimes(self, dateCol, timeCol):
        timeSeries = (dateCol.astype(str) + " " + timeCol.astype(str))
        if not timeSeries.is_unique:
            dupeTimes = timeSeries.value_counts()[timeSeries.value_counts() > 1].index.tolist()
            for time in dupeTimes:
                self.collectError(self.ERRORS['nonUniqueTime'], time)
    
    def findDataSheet(self, filename) -> int:
        """ Find the index of the first XLSX sheet with suitable data columns """
        sheets = pd.read_excel(filename, None, keep_default_na=False)
        for S in sheets:
            if set(sheets[S].columns).issuperset(self.EXPECTED_COLS):
                return sheets[S]
        self.collectError(self.ERRORS['missingColumn'], str(self.EXPECTED_COLS))
        exit(0)

    def checkFile(self, filename):
        fileExt = pathlib.Path(filename).suffix
        self.log.info(f"\nChecking: {pathlib.Path(filename).name}")
        if fileExt == ".csv":
            self.filetype = 'text'
            df = pd.read_csv(filename, keep_default_na=False, float_precision="round_trip")
        elif fileExt in [".xls", ".xlsx"]:
            self.filetype = 'excel'
            df = self.findDataSheet(filename)
        else:
            self.collectError(self.ERRORS['unsupportedExt'], fileExt)
            return 
       
        # Checks
        
        # Non-template/non-depth column check
        dropCols = self.checkExtraColNames(df.columns)
        df.drop(columns=dropCols, inplace=True)
        
        # Check all of the column names
        self.checkColNames(df.columns)
            
        # Latitude and longitude check
        df.apply(lambda row: self.checkLatitude(row['latitude'], row.name), axis=1)
        df.apply(lambda row: self.checkLongitude(row['longitude'], row.name), axis=1)
        
        # Time is in xx:xx:xx format
        if self.filetype == "excel":
            self.log.info("Unable to check time format for Excel files because of how times are stored in Excel. Try converting to CSV.")
        else:
            df.apply(lambda row: self.checkTimeFmt(row['time_HH:MM:SS'], row.name), axis=1)
        
        # Date is in yyyy-mm-dd format
        if self.filetype == "excel":
            self.log.info("Unable to check date format for Excel files because of how dates are stored in Excel. Try converting to CSV.")
        else:
            df.apply(lambda row: self.checkDateFmt(row['date_YYYY-MM-DD'], row.name), axis=1)
        
        for i in range(len(self.EXPECTED_COLS), len(df.columns)): # All other metre columns
            # Temperature can be converted to float or not
            df.apply(lambda row: self.checkTemperatures(row[df.columns[i]], row.name), axis=1)
        
        # Check if all time values are unique
        self.checkUniqueTimes(df["date_YYYY-MM-DD"], df["time_HH:MM:SS"])

        self.log.info("Done")

    @staticmethod
    def pathExists(path):
        if pathlib.Path(path).exists():
            return path
        else:
            raise argparse.ArgumentTypeError(f"The path specified does not exist: {path}")

    @staticmethod
    def createPathIfExists(path):
        if not pathlib.Path(path).exists():
            pathlib.Path(path).mkdir(parents=True, exist_ok=True)
        return path

    def checkPath(self, pathLoc, isVerbose, logPath):
        pathLoc = pathlib.Path(pathLoc)
        finalFilePath = ''
        f_handle = None
        
        self.log.setLevel(logging.INFO)
        
        # Verbose printing or not
        if isVerbose:
            console = logging.StreamHandler()
            console.setLevel(logging.INFO)
            self.log.addHandler(console)
        
        # Concerning log file output
        if logPath is not None:
            if not pathlib.Path(logPath).suffix:
                # If there is no suffix in the path, it means a directory was given
                FileDataChecker.createPathIfExists(logPath)
            else:
                # There is a parent folder that needs to be made, a file w/ extension was given
                FileDataChecker.createPathIfExists(pathlib.Path(logPath).parent)
        
        currTime = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H-%M-%S')
        if logPath:
            logPathObj = pathlib.Path(logPath)
            # Set logging file path
            if logPathObj.is_dir():
                # Is a directory - add file name and current time to log name
                finalFilePath = str(logPathObj.joinpath(pathLoc.stem)) + '_' + currTime + '.txt'
                f_handle = logging.FileHandler(filename=finalFilePath, mode='w')
            else:
                # User specified filename
                finalFilePath = logPathObj
                f_handle = logging.FileHandler(filename=finalFilePath, mode='w')
        else:
            # No log path specified, create in same directory as this script
            finalFilePath = pathLoc.stem + '_' + currTime + '.txt'
            f_handle = logging.FileHandler(filename=finalFilePath, mode='w')
            
        # File handler always writes to a file
        f_handle.setFormatter(logging.Formatter())
        f_handle.setLevel(logging.INFO)
        self.log.addHandler(f_handle)
        self.log.info(currTime)
        
        # If it's a zip file, unzip it to a directory and set the path to that directory.
        zipDir = ''
        if pathLoc.suffix == ".zip":
            # Current time appended to prevent read of similarly named zips
            zipDir = str(pathLoc.parent) + "\\" + str(pathLoc.stem) + currTime
            # Unzip the archive - expected zip files only
            with zip.ZipFile(pathLoc) as zf:
                zf.extractall(zipDir)
            pathLoc = pathlib.Path(zipDir)
            
        if pathLoc.is_dir():
            for filename in pathLoc.iterdir():
                if filename.is_file():
                    self.checkFile(filename)
            # Delete the unzipped directory and contents afterwards
            if zipDir != '':
                for filename in pathLoc.iterdir():
                    pathlib.Path(filename).unlink()
                pathlib.Path(zipDir).rmdir()
        elif pathLoc.is_file():
            filename = pathLoc
            self.checkFile(filename)
        else:
            self.log.error("Error: unrecognized path")
            return -1
        return 1

    @property
    def filetype(self):
        return self._filetype

    @filetype.setter
    def filetype(self, value):
        if value in self.FILETYPES:
            self._filetype = value
        else:
            raise ValueError("Not a valid filetype")

if __name__ == "__main__":
    # Parser init
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''
        Conducts checks on column names and values for CSV, XLS, and XLSX NTGS-standard data files. XLS/XLSX files have been known to contain problematic values themselves, consider converting to CSV.
        
        * Only checks the first valid sheet of an excel file
        * Expects data files to contain the first 6 columns with the exact names:
        ***  project_name
        ***  site_id
        ***  latitude
        ***  longitude
        ***  date_YYYY-MM-DD
        ***  time_HH:MM:SS
        
        Error Explanations:
        Invalid Time - Time does not follow a valid time in the format HH:MM:SS.
        Invalid Date - Date values should be formatted as YYYY-MM-DD.
        Non Unique Time - The combination of date and time column values should uniquely occur/never repeat.
        Column Order - The first 6 column names are not in the correct order.
        Extra Column - A column that is not a part of the expected 6 column names and is not a valid depth column is found.
        No Measurements - No measurement columns are detected in the file.
        File Type - The file read in is not supported.
        Coordinate - A latitude or longitude value contains something that is not valid.
        Latitude - A latitude value is found that is not valid (Less than -90 or greater than 90).
        Longitude - A longitude value is found that is not valid (Less than -180 or greater than 180).
        Temperature - A temperature value is found that is not a valid temperature.
        '''))
    parser.add_argument('location', metavar='path', type=FileDataChecker.pathExists, help='Path to specified directory or single file. Non-recursive if directory.')
    parser.add_argument('-v', '--verbose', action='store_true', help='Displays detected issues with the checked file during operation of the program.')
    parser.add_argument('--log-path', metavar='log-path', help='Specified path where output log file should be saved to.')
    arguments = vars(parser.parse_args())
    
    checkingFilePath = arguments['location']
    isVerbose = arguments['verbose']
    outputLogPath = arguments['log_path']
    
    fdc = FileDataChecker()
    fdc.checkPath(checkingFilePath, isVerbose, outputLogPath)



import argparse
import datetime
import numpy as np
import pandas as pd
import pathlib
import re
import textwrap
from typing import Union, Optional

class CSVMelter():
    """ Class to convert NTGS csv from long to wide format"""
    def __init__(self, areaTimezone: Union[float,int] = -7):
        self.areaTimezone = datetime.timezone(datetime.timedelta(hours=areaTimezone))

    @staticmethod
    def timezone_check(tz: Union[float, int, str]) -> float:
        """ Ensure timezone is proper """
        tz = float(tz)
        if (tz < -15) or (tz > 14):
            raise argparse.ArgumentTypeError(f"Timezone value {tz} out of bounds!")
        return tz
        
    @staticmethod
    def pathExists(path: Union[str, pathlib.Path]) -> Union[str, pathlib.Path]:
        """ Check if path exists """
        if pathlib.Path(path).exists():
            return path
        else:
            raise argparse.ArgumentTypeError(f"The path specified does not exist: {path}")

    def getISOFormat(self, date: Union[str, datetime.datetime], time: Union[str, datetime.datetime]) -> str:
        """ Convert date and time into ISO-formatted datetime """
        if isinstance(date, str):
            # Case from CSV files where the date is string
            try:
                year, month, day = [int(dateVal) for dateVal in date.split("-")]
            except ValueError:
                raise ValueError(f"The date {date} was unable to be parsed. The format required is YYYY-MM-DD.")
        elif isinstance(date, datetime.datetime):
            # Case XLSX files - are "timestamp" objects
            year, month, day = date.year, date.month, date.day
        else:
            raise ValueError(f"The date {date} was unable to be parsed.")
            
        if isinstance(time, str):
            try:
                h, m, s = [int(timeVal) for timeVal in time.split(":")]
            except ValueError:
                raise ValueError(f"The time {time} was unable to be parsed. The format required is (H)H:MM:SS.")
        elif isinstance(time, datetime.time):
            h, m, s = int(time.hour), time.minute, time.second
        else:
            raise ValueError(f"The time {time} was unable to be parsed.")
        return datetime.datetime(year,month,day,hour=h,minute=m,second=s,tzinfo=self.areaTimezone).isoformat()
    
    def meltDataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """ Convert 'wide' dataframe to 'long' style """
        originalCols = []
        meltCols = []
                    
        # Regex detect columns 
        for colName in df.columns:
            # if colName ends with '_m', it's depth
            if "_m" in colName:
                obj = re.search("^(?:(?!.*)|[+-]?((\d+\.?\d*)|(\.\d+)))(_m)$", colName)
                if obj is not None:
                    meltCols.append(colName)
                else:
                    raise ValueError(f"The column {colName} cannot be converted to a valid metre value. Ensure that all columns are free of errors.")

        tmp = set(meltCols)
        originalCols = [x for x in df.columns if x not in tmp]

        if len(meltCols) <= 0:
            raise ValueError("No depth columns found.")

        newColName = "depth"
        newValueName = "temperature"

        # Transpose of the columns occurs here
        melted_df = pd.melt(df, id_vars=originalCols, value_vars=meltCols, var_name=newColName, value_name=newValueName)
       
        # Convert depth values to floats and truncate the "_m"
        melted_df['depth'] = melted_df['depth'].apply(lambda val: float(val[:-2]))

        # Reorder the columns
        melted_df = melted_df[['project_name', 'site_id','latitude','longitude','time','depth','temperature']]
        
        return melted_df
    
    def parseTimeFormat(self, time: Union[str, datetime.datetime]) -> "tuple[str, str]":
        """ Split datetime into date, time """
        try:
            timeObj = datetime.datetime.strptime(str(time), "%Y-%m-%d %H:%M:%S")
        except ValueError as e:
            raise ValueError(f"The time {time} was unable to be interpreted in the YYYY-MM-DD HH:MM:SS format.")
        return timeObj.strftime("%Y-%m-%d"), timeObj.strftime("%H:%M:%S")
    
    def stringifyDepths(self, columns: list) -> "tuple[dict, list]":
        """ """
        depthVals = [name for name in columns if type(name) in (float, int)]
        depthDict = {}
        depths = []
        for name in depthVals:
            if not type(name) == str:
                # Check if you can change depth into int without losing info, otherwise keep as float, then append _m
                depthDict[name] = f"{str(int(name))}_m" if int(name) == name and isinstance(name, float) else f"{str(name)}_m"
                depths.append(f"{str(int(name))}_m" if int(name) == name and isinstance(name, float) else f"{str(name)}_m")
        return depthDict, depths
    
    def longToWide(self, df: pd.DataFrame) -> pd.DataFrame:
        """ Convert 'long' dataframe to 'wide' dataframe """
        depths = df.groupby(["time", "depth"])["temperature"].aggregate('mean').unstack()
        depths = depths.reset_index()
        wide = df[["project_name", "site_id", "latitude", "longitude"]].copy()
        wide = wide.join(depths, how="right")
        for col in ["project_name", "site_id", "latitude", "longitude"]:
            wide[col] = wide[col].mode(dropna=True)[0]
        wide["date_YYYY-MM-DD"], wide["time_HH:MM:SS"] = zip(*wide["time"].apply(self.parseTimeFormat))
        colOrder = ['project_name', 'site_id', 'latitude', 'longitude', 'date_YYYY-MM-DD', 'time_HH:MM:SS'] 
        depthDict, strDepths = self.stringifyDepths(wide.columns.tolist())
        wide.rename(columns=depthDict, inplace=True)
        colOrder.extend(strDepths)
        wide = wide[colOrder]
        return wide
    
    def getMeltedDataframe(self, filename: str) -> pd.DataFrame:
        """ Get melted dataframe from a file"""
        if pathlib.Path(filename).suffix == ".csv":
            try:
                df = pd.read_csv(filename, keep_default_na=False, float_precision="round_trip", parse_dates={"time": [4,5]}, date_parser=self.getISOFormat)
            except IndexError:
                raise IndexError("There are insufficient columns, the file format is invalid.")
        elif pathlib.Path(filename).suffix in [".xls", ".xlsx"]:
            try:
                df = pd.read_excel(filename, keep_default_na=False, parse_dates={"time": [4,5]}, date_parser=self.getISOFormat)
            except IndexError:
                raise IndexError("There are insufficient columns, the file format is invalid.") 
        else:
            raise TypeError("Unsupported file extension.")
            
        melted_df = self.meltDataframe(df)

        return melted_df

    def meltFile(self, filename: str, outLoc:str = '.') -> int:
        """ convert 'wide' csv file to 'long' csv file """
        melted_df = self.getMeltedDataframe(filename)

        fs = f"melted_{pathlib.Path(filename).stem}.csv"
        if outLoc != '.' and outLoc is not None: 
            fs = pathlib.Path(outLoc, f"melted_{pathlib.Path(filename).stem}.csv")
        else:
            pass
        melted_df.to_csv(fs, encoding="utf-8", index=False)
        return 1
        
if __name__ == '__main__':
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''
        Converts CSV, XLS, or XLSX files in the NTGS-style "wide" format to CSV files in "long" format by creating a new column for depth.
        The file being transformed must be free of errors prior to running this.
        '''))
    parser.add_argument('location', metavar='path', type=CSVMelter.pathExists, help='Path to file')
    parser.add_argument('--output-path', metavar='output-path', type=CSVMelter.pathExists, help='Optional directory of the melted CSV file to write to')
    parser.add_argument('--timezone-offset', metavar='timezone-offset', type=CSVMelter.timezone_check, help='Optional value indicating timezone offset from UTC-0. If no value is given, timezone defaults to UTC-7. To represent half hour, use "-3.5" for Newfoundland time as an example.')
    arguments = vars(parser.parse_args())
    cm = None
    if arguments['timezone_offset'] is not None:
        cm = CSVMelter(arguments['timezone_offset'])
    else:
        cm = CSVMelter()
    cm.meltFile(arguments['location'], arguments['output_path'])



from abc import ABC, abstractmethod
from netCDF4 import Dataset


class MetadataHandler(ABC):

    @abstractmethod
    def write_attributes(nc_file: Dataset, loc:str) -> None:
        """ Write metadata to attributes in netcdf file for location """
        pass

    @abstractmethod
    def write_variables(nc_file: Dataset, loc:str) -> None:
        """ Write metadata to variables in netcdf file for location """
        pass

    def write_global_attribute(self, nc_file: Dataset, key, value, overwrite=False) -> None:
        with Dataset(nc_file) as dataset:
            if hasattr(dataset, key) and not overwrite:
                raise ValueError(f"Attribute {key} already exists for dataset {nc_file}")
            else:
                dataset.setncattr(key, value)





from netCDF4._netCDF4 import Variable


variables = {"template": {"standard_name": "",
                          "standard_name_url": "",
                          "long_name": "",
                          "units": "",
                          "ioos_category": "",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

             "excess_ice":{"standard_name": "ice_volume_in_frozen_ground_in_excess_of_pore_volume_in_unfrozen_ground_expressed_as_fraction_of_frozen_ground_volume",
                          "standard_name_url": "https://mmisw.org/ont/cf/parameter/ice_volume_in_frozen_ground_in_excess_of_pore_volume_in_unfrozen_ground_expressed_as_fraction_of_frozen_ground_volume",
                          "long_name": "Excess ice content",
                          "units": "percent",
                          "valid_min":"0",
                          "ioos_category": "Ice Distribution",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "sand": {"standard_name": "mass_fraction_of_sand_in_soil",
                          "standard_name_url": "https://mmisw.org/ont/cf/parameter/mass_fraction_of_sand_in_soil",
                          "long_name": "Sand content (by mass)",
                          "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                          "ioos_category": "Other",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "silt": {"standard_name": "mass_fraction_of_silt_in_soil",
                          "standard_name_url": "https://mmisw.org/ont/cf/parameter/mass_fraction_of_silt_in_soil",
                          "long_name": "Silt content (by mass)",
                          "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                          "ioos_category": "Other",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "clay": {"standard_name": "mass_fraction_of_clay_in_soil",
                          "standard_name_url": "https://mmisw.org/ont/cf/parameter/mass_fraction_of_clay_in_soil",
                          "long_name": "Clay content (by mass)",
                          "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                          "ioos_category": "Other",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},
            
            "gravel": {"standard_name": "mass_fraction_of_gravel_in_soil",
                          "standard_name_url": "https://mmisw.org/ont/cf/parameter/mass_fraction_of_gravel_in_soil",
                          "long_name": "Gravel content (by mass)",
                          "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                          "ioos_category": "Other",
                          "accuracy": "",
                          "precision": "",
                          "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "liquid_limit": {#"standard_name": "soil_liquid_limit",
                            #"standard_name_url": "https://mmisw.org/ont/cf/parameter/soil_liquid_limit",
                            "long_name": "Atterberg liquid limit",
                            "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                            "ioos_category": "Other",
                            "accuracy": "",
                            "precision": "",
                            "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "plastic_limit": {#"standard_name": "soil_plastic_limit",
                             #"standard_name_url": "https://mmisw.org/ont/cf/parameter/soil_plastic_limit",
                             "long_name": "Atterberg plastic limit",
                             "units": "percent",
                          "valid_min":"0",
                          "valid_max":"100",
                             "ioos_category": "Other",
                             "accuracy": "",
                             "precision": "",
                             "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "ground_ice_class": {"long_name": "Ground ice classification",
                                "ioos_category": "Other",
                                "accuracy": "",
                                "precision": "",
                                "_Encoding": "UTF-8",
                                "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "soil_ph": {"standard_name": "soil_water_ph",
                             "standard_name_url": "https://mmisw.org/ont/cf/parameter/soil_water_ph",
                             "long_name": "Soil water pH",
                                # "units": "1",
                                "ioos_category": "Other",
                                "accuracy": "",
                                "precision": "",
                                "_Encoding": "UTF-8",
                                "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},
            
            "materials": {#"standard_name": "",
                          #"standard_name_url": "",
                             "long_name": "Material classification",
                                #"units": "1",
                                "ioos_category": "Other",
                                "accuracy": "",
                                "precision": "",
                                "_Encoding": "UTF-8",
                                "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "materials_description": {#"standard_name": "",
                          #"standard_name_url": "",
                             "long_name": "Material description",
                                #"units": "1",
                                "ioos_category": "Other",
                                "accuracy": "",
                                "precision": "",
                                "_Encoding": "UTF-8",
                                "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "volumetric_moisture": {"standard_name": "volume_fraction_of_condensed_water_in_soil",
                                    "standard_name_url": "https://mmisw.org/ont/cf/parameter/volume_fraction_of_condensed_water_in_soil",
                                  "long_name": "Volumetric moisture content",
                                  "units": "percent",
                                  "valid_min" : "0",
                                "ioos_category": "Other",
                                "accuracy": "",
                                "precision": "",
                                "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "gravimetric_moisture": {#"standard_name": "",
                                     #"standard_name_url": "",
                                     "long_name": "Gravimetric moisture content",
                                     "units": "percent",
                                     "valid_min":"0",
                                     "ioos_category": "Other",
                                     "accuracy": "",
                                     "precision": "",
                                     "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},

            "visible_ice": {"standard_name": "volume_fraction_of_frozen_water_in_soil",
                            "standard_name_url": "https://mmisw.org/ont/cf/parameter/volume_fraction_of_frozen_water_in_soil",
                            "long_name": "Visible ice content",
                            "units": "percent",
                            "valid_min":"0",
                            "ioos_category": "Other",
                            "accuracy": "",
                            "precision": "",
                            "coordinates": "depth_below_ground_surface bottom_of_interval top_of_interval time latitude longitude platform_name profile"},
    }

def add_variable(rootgrp, name, dtype, dimensions, fill_value=None) -> Variable:
    if name not in variables.keys():
        raise KeyError("Variable name not in variable templates list")
    else:
        vt = variables[name]

    var = rootgrp.createVariable(name, dtype, dimensions, fill_value=fill_value)
    
    for key, val in vt.items():
        var.setncattr(key, val)

    return var


from pfit.FileDataChecker import FileDataChecker
from pfit.CSVColMelter import CSVMelter
from pfit.TabularMetadataHandler import TabularMetadataHandler
from pfit.ErddapXmlGenerator import ErddapXmlGenerator
from pfit.NTGS_to_NetCDF import NtgsThermalDataset
from pfit.NTGSTemperatureMetadataHandler import NTGSTemperatureMetadataHandler
